#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <inttypes.h>
#include <errno.h>
#include <signal.h>
#include <assert.h>
#include <fcntl.h>
#include <setjmp.h>
#include <pthread.h>
#include <sys/time.h>
#include <sys/mman.h>
#include <uuid/uuid.h>

#include "msr-index.h"
#include "xenner.h"
#include "mm.h"

/* ------------------------------------------------------------------ */

static int shutdown;
static int termsig;

struct stats {
    int cb_out;
    int cb_run;
    int cb_hlt;
} cs, ls;

/* ------------------------------------------------------------------ */

static int shutdown;
static int termsig;

static sigset_t main_signals;

static void sighandle(int sig)
{
    switch (sig) {
    case SIGHUP:
	/* ignore */
	break;
    case SIGINT:
	if (!shutdown) {
	    shutdown++;
	    break;
	}
	/* else fall through */
    case SIGTERM:
    case SIGQUIT:
    default:
	fprintf(stderr, "%s: %s\n", __FUNCTION__, strsignal(sig));
	termsig = sig;
    }
}

static void sigsetup(struct xenvm *xen)
{
    struct sigaction act, old;

    /* install handlers */
    memset(&act,0,sizeof(act));
    act.sa_handler = sighandle;
    sigemptyset(&act.sa_mask);
    sigaction(SIGHUP,  &act, &old);
    sigaction(SIGINT,  &act, &old);
    sigaction(SIGTERM, &act, &old);
    sigaction(SIGQUIT, &act, &old);

    sigemptyset(&main_signals);
    sigaddset(&main_signals, SIGINT);
    sigaddset(&main_signals, SIGTERM);
    sigaddset(&main_signals, SIGQUIT);
}

void alive(void)
{
    static const char alive[] = ".oOo";
    static const int chars = sizeof(alive)/sizeof(alive[0]) -1;
    static int count;
    
    fprintf(stderr, "[%c]\r", alive[count++ % chars]);
}

uint64_t get_systime(void)
{
    uint64_t systime;
    struct timeval tv;

    gettimeofday(&tv, NULL);
    systime  = tv.tv_sec  * (uint64_t)1000000000;
    systime += tv.tv_usec * (uint64_t)1000;
    return systime;
}

static int balloon(struct xenvm *xen)
{
    uint32_t page;
    int advice;

    /* Hmm, keep track of state and signal only changes?  Not sure it is
     * worth the effort given ballooning is an infrequent operation in
     * Xen due to being fully admin-controlled (aka manual) */
    for (page = 0; page < xen->pg_guest; page++) {
	advice = 0;
	if (xen->m2p_32)
	    if (xen->m2p_32[page] == (uint32_t)INVALID_M2P_ENTRY)
		advice = MADV_DONTNEED;
	if (xen->m2p_64)
	    if (xen->m2p_64[page] == (uint64_t)INVALID_M2P_ENTRY)
		advice = MADV_DONTNEED;
	if (advice)
	    madvise(xen->memory + page * PAGE_SIZE, PAGE_SIZE, advice);
    }
    return 0;
}

/* ------------------------------------------------------------------ */

void update_time(struct xenvcpu *vcpu)
{
    struct xenvm *xen = vcpu->vm;
    uint64_t systime;

    vcpu->systime = get_systime();
    systime = vcpu->systime - vcpu->vm->boot;

    d2printf("systime %" PRId64 ".%09" PRId64 "\n",
	     systime / 1000000000, systime % 1000000000);

    if (vcpu->xentime) {
	vcpu->xentime->version++;
	vcpu->xentime->system_time       = systime;
	vcpu->xentime->tsc_timestamp     = get_msr(vcpu, MSR_IA32_TSC);
	vcpu->xentime->tsc_to_system_mul = xen->tsc_mul_frac;
	vcpu->xentime->tsc_shift         = xen->tsc_shift;
	vcpu->xentime->version++;
    }

#if 1 /* FIXME: move elsewhere */
    xen->xen_wc->wc_version++;
    xen->xen_wc->wc_sec  = xen->boot / 1000000000;
    xen->xen_wc->wc_nsec = xen->boot % 1000000000;
    xen->xen_wc->wc_version++;
#endif
}

int raise_event(struct xenvcpu *vcpu, int port)
{
    struct xenvm *xen = vcpu->vm;

    kvm_set_irq_level(xen->kvm, port, 1);
    return 0;
}

int xenner_register_cb(struct xenvm *xen, enum xenner_cb_type type,
		       char *name, int handle, void *cb_data,
		       int (*callback)(struct xenvm *xen, int fd, void *data))
{
    struct xenner_callback *cb;
    long flags;

    cb = malloc(sizeof(*cb));
    memset(cb,0,sizeof(*cb));
    cb->handle   = handle;
    cb->name     = name;
    cb->cb_data  = cb_data;
    cb->callback = callback;
    list_add_tail(&cb->list, &xen->callbacks[type]);

    switch (type) {
    case XENNER_CB_IOFD:
	/* set nonblocking mode */
	flags  = fcntl(cb->handle, F_GETFL, 0);
	flags |= O_NONBLOCK;
	if (0 != fcntl(cb->handle, F_SETFL, flags)) {
	    perror("fcntl F_SETFL");
	    return -1;
	}
	break;
    default:
	/* nothing special to do */
	break;
    }
    return 0;
}

int xenner_callback_evtchn(struct xenvm *xen, int port)
{
    struct list_head *item;
    struct xenner_callback *cb;

    list_for_each(item, &xen->callbacks[XENNER_CB_EVTCHN]) {
	cb = list_entry(item, struct xenner_callback, list);
	if (cb->handle != port)
	    continue;
	d2printf("%s: calling %s\n", __FUNCTION__, cb->name);
	cb->callback(xen, cb->handle, cb->cb_data);
	return 0;
    }
    return -1;
}

static void check_events(struct xenvm *xen)
{
    if (shutdown) {
	static int once = 1;
	if (once) {
	    xenstore_shutdown(xen, "halt");
	    once = 0;
	}
    }
    
    /* balloon: check for work
     * FIXME: find better place */
    if (xen->vminfo->balloon) {
	xen->vminfo->balloon = 0;
	balloon(xen);
    }
}

/* ------------------------------------------------------------------ */

void need_regs(struct xenvcpu *vcpu)
{
    if (vcpu->regs_up2date)
	return;
    kvm_get_regs(vcpu->vm->kvm, vcpu->id, &vcpu->regs);
    vcpu->regs_up2date = 1;
}

void need_sregs(struct xenvcpu *vcpu)
{
    if (vcpu->sregs_up2date)
	return;
    kvm_get_sregs(vcpu->vm->kvm, vcpu->id, &vcpu->sregs);
    vcpu->sregs_up2date = 1;
}

void flush_regs(struct xenvcpu *vcpu)
{
    assert(vcpu->regs_up2date);
    vcpu->regs_dirty = 1;
}

void flush_sregs(struct xenvcpu *vcpu)
{
    assert(vcpu->sregs_up2date);
    vcpu->sregs_dirty = 1;
}

void xenner_cleanup(struct xenvm *xen, int stats)
{
    /* going down now ... */
    if (xen->vminfo)
	xen->vminfo->dying = 1;

    if (stats) {
	section_print(xen, __FUNCTION__, "statistics");
	hypercall_stats(xen);
    }

    section_print(xen, __FUNCTION__, "cleaning up");
    xenstore_fini(xen);
    evtchn_fini(xen);
    qemu_cleanup(xen);
    vmcore_fini(xen);
}

/* ------------------------------------------------------------------ */

static void print_stack_32(struct xenvm *xen, char *name, uint32_t esp, uint32_t *ptr)
{
    int i;
    
    logprintf(xen, "--- %s stack trace ---\n", name);
    for (i = 0; i < 128; i++) {
	if (0 == addr_offset(esp + 4*i))
	    break;
	if (0 == (i % 4))
	    logprintf(xen, "%08" PRIx32 ":", esp + 4*i);
	logprintf(xen, " %08" PRIx32, ptr[i]);
	if (3 == (i % 4))
	    logprintf(xen, "\n");
    }
    if (0 != (i % 4))
	logprintf(xen, "\n");
}

static void print_stack_64(struct xenvm *xen, char *name, uint64_t esp, uint64_t *ptr)
{
    int i;
    
    logprintf(xen, "--- %s stack trace ---\n", name);
    for (i = 0; i < 128; i++) {
	if (0 == addr_offset(esp + 8*i))
	    break;
	if (0 == (i % 4))
	    logprintf(xen, "%016" PRIx64 ":", esp + 8*i);
	logprintf(xen, " %016" PRIx64, ptr[i]);
	if (3 == (i % 4))
	    logprintf(xen, "\n");
    }
    if (0 != (i % 4))
	logprintf(xen, "\n");
}

void print_guest_stack(struct xenvcpu *vcpu, uint64_t rsp)
{
    if (vcpu->vm->mode == XENMODE_64) {
	uint64_t *data = guest_vaddr_to_ptr(vcpu, rsp);
	if (data)
	    print_stack_64(vcpu->vm, "guest64", rsp, data);
    } else {
	uint32_t *data = guest_vaddr_to_ptr(vcpu, rsp);
	if (data)
	    print_stack_32(vcpu->vm, "guest32", rsp, data);
    }
}

void print_emu_stack(struct xenvm *xen, uint64_t rsp)
{
    if (xen->mode == XENMODE_64) {
	uint64_t *data = emu_vaddr_to_ptr(xen, rsp);
	if (data)
	    print_stack_64(xen, "emu64", rsp, data);
    } else {
	uint32_t *data = emu_vaddr_to_ptr(xen, rsp);
	if (data)
	    print_stack_32(xen, "emu32", rsp, data);
    }
}

static void dump_vcpu(struct xenvcpu *vcpu, uint64_t guest_rsp)
{
    struct xenvm *xen = vcpu->vm;

    if (!pthread_equal(pthread_self(), vcpu->thread)) {
	/* deadlock risk, vcpu might be running in another thrad ... */
	logprintf(xen, "--- skipping vcpu %d register dump ---\n", vcpu->id);
	return;
    }

    logprintf(xen, "--- registers (vcpu %d)---\n", vcpu->id);
    kvm_show_regs(xen->kvm, vcpu->id);
    print_msrs(vcpu);

    need_regs(vcpu);
    if (vcpu->regs.rsp > xen->emu_vs && vcpu->regs.rsp < xen->emu_ve)
	print_emu_stack(xen, vcpu->regs.rsp);
    else if (!guest_rsp && vcpu->regs.rsp)
	guest_rsp = vcpu->regs.rsp;

    if (guest_rsp)
	print_guest_stack(vcpu, guest_rsp);	
}

void vm_kill(struct xenvcpu *vcpu, const char *reason, uint64_t guest_rsp)
{
    struct xenvm *xen = vcpu->vm;
    
    banner_print(xen, reason);
    dump_vcpu(vcpu, guest_rsp);
    xenner_cleanup(xen, 1);
    exit(1);
}

/* ------------------------------------------------------------------ */
/* functions                                                          */

static int cb_inb(void *opaque, uint16_t addr, uint8_t *value)
{
    struct xenvm *xen = opaque;

    d2printf("io: %s, addr 0x%" PRIx16 "\n",
	     __FUNCTION__, addr);
    *value = 0xff;
    return 0;
}

static int cb_inw(void *opaque, uint16_t addr, uint16_t *value)
{
    struct xenvm *xen = opaque;

    d2printf("io: %s, addr 0x%" PRIx16 "\n",
	     __FUNCTION__, addr);
    *value = 0xffff;
    return 0;
}

static int cb_inl(void *opaque, uint16_t addr, uint32_t *value)
{
    struct xenvm *xen = opaque;

    d2printf("io: %s, addr 0x%" PRIx16 "\n",
	     __FUNCTION__, addr);
    *value = 0xffffffff;
    return 0;
}

static int cb_outb(void *opaque, uint16_t addr, uint8_t value, int vcpu_id)
{
    struct xenvm *xen = opaque;
    struct xenvcpu *vcpu = xen->vcpu + vcpu_id;
    char *msg;

    cs.cb_out++;
    switch (addr) {
    case 0xe0: /* hypercall */
	do_xen_hypercall(vcpu, value);
	break;

    case 0xe1: /* vmexit nop() */
	break;

    case 0xea: /* emu: print string */
	need_regs(vcpu);
	msg = emu_vaddr_to_ptr(xen, vcpu->regs.rdx);
	conemu_print(vcpu, msg);
	break;

    case 0xeb: /* emu: panic */
	vm_kill(vcpu, "emu: panic()", 0);
	break;

    case 0xec: /* debug trap */
	do_debug_trap(vcpu);
	break;

    case 0xed: /* shutdown */
	logprintf(xen, "guest requests shutdown, exiting.\n");
	xenner_cleanup(xen, 1);
	exit(0);

    case 0xee: /* get evtchn port */
	need_regs(vcpu);
	vcpu->regs.rax = evtchn_port(xen, "irqchip");
	flush_regs(vcpu);
	break;

    default:
	d2printf("io: %s, addr 0x%" PRIx16 " value 0x%" PRIx8 "\n",
		__FUNCTION__, addr, value);
	break;
    }
    return 0;
}

static int cb_outw(void *opaque, uint16_t addr, uint16_t value)
{
    struct xenvm *xen = opaque;

    d1printf("io: %s, addr 0x%" PRIx16 " value 0x%" PRIx16 "\n",
	     __FUNCTION__, addr, value);
    return 0;
}

static int cb_outl(void *opaque, uint16_t addr, uint32_t value)
{
    struct xenvm *xen = opaque;

    d1printf("io: %s, addr 0x%" PRIx16 " value 0x%" PRIx32 "\n",
	     __FUNCTION__, addr, value);
    return 0;
}

static int cb_debug(void *opaque, int vcpu_id)
{
    struct xenvm *xen = opaque;
    
    d3printf("callback: %s/%d\n", __FUNCTION__, vcpu_id);
    return 0;
}

static int cb_halt(void *opaque, int vcpu_id)
{
    struct xenvm *xen = opaque;
    struct xenvcpu *vcpu = xen->vcpu + 0;

    /* should not be called any more */
    vm_kill(vcpu, __FUNCTION__, 0);
    return 0;
}

static int cb_shutdown(void *opaque, int vcpu_id)
{
    struct xenvm *xen = opaque;
    struct xenvcpu *vcpu = xen->vcpu + vcpu_id;

    d3printf("callback: %s/%d\n", __FUNCTION__, vcpu_id);
    vm_kill(vcpu, "crashed", 0);
    return 0;
}

static int cb_io_window(void *opaque)
{
    struct xenvm *xen = opaque;

    d2printf("callback: %s\n", __FUNCTION__);
    return 0;
}

static int cb_try_push_interrupts(void *opaque)
{
    struct xenvm *xen = opaque;
    struct xenvcpu *vcpu = xen->vcpu + 0;

    /* should not be called any more */
    vm_kill(vcpu, __FUNCTION__, 0);
    return 0;
}

static void cb_post_kvm_run(void *opaque, int vcpu_id)
{
    struct xenvm *xen = opaque;

    pthread_mutex_lock(&xen->biglock);
    d3printf("callback: %s/%d\n", __FUNCTION__, vcpu_id);
}

static int cb_pre_kvm_run(void *opaque, int vcpu_id)
{
    struct xenvm *xen = opaque;
    struct xenvcpu *vcpu = xen->vcpu + vcpu_id;

    cs.cb_run++;
    d3printf("callback: %s/%d\n", __FUNCTION__, vcpu_id);

    update_time(vcpu);

    /* registers */
    if (vcpu->regs_dirty) {
	kvm_set_regs(xen->kvm, vcpu->id, &vcpu->regs);
	vcpu->regs_dirty = 0;
    }
    if (vcpu->sregs_dirty) {
	kvm_set_sregs(xen->kvm, vcpu->id, &vcpu->sregs);
	vcpu->sregs_dirty = 0;
    }
    vcpu->regs_up2date  = 0;
    vcpu->sregs_up2date = 0;

    pthread_mutex_unlock(&xen->biglock);
    return 0;
}

/* ------------------------------------------------------------------ */

static struct kvm_callbacks xenner_cb = {
    .inb                 = cb_inb,
    .inw                 = cb_inw,
    .inl                 = cb_inl,
    .outb                = cb_outb,
    .outw                = cb_outw,
    .outl                = cb_outl,
    .debug               = cb_debug,
    .halt                = cb_halt,
    .shutdown            = cb_shutdown,
    .io_window           = cb_io_window,
    .try_push_interrupts = cb_try_push_interrupts,
    .post_kvm_run        = cb_post_kvm_run,
    .pre_kvm_run         = cb_pre_kvm_run,
};

static struct kvm_para_features {
        int cap;
        int feature;
} para_features[] = {
#if 0 /* FIXME: mask out, not working yet */
        { KVM_CAP_CLOCKSOURCE, KVM_FEATURE_CLOCKSOURCE },
#endif
        { KVM_CAP_NOP_IO_DELAY, KVM_FEATURE_NOP_IO_DELAY },
        { KVM_CAP_PV_MMU, KVM_FEATURE_MMU_OP },
#ifdef KVM_CAP_CR3_CACHE
        { KVM_CAP_CR3_CACHE, KVM_FEATURE_CR3_CACHE },
#endif
};

int kvmbits_init(struct xenvm *xen)
{
    int i;
    
    /* init kvm */
    xen->kvm = kvm_init(&xenner_cb, xen);
    if (!xen->kvm) {
	logprintf(xen, "kvm_init failed\n");
	goto err;
    }

    if (!kvm_check_extension(xen->kvm, KVM_CAP_SET_TSS_ADDR)) {
	logprintf(xen, "Error: KVM_CAP_SET_TSS_ADDR is not available.\n");
	logprintf(xen, "Most likely your kvm version is too old.\n");
	logprintf(xen, "Minimum requirement is kvm-49 (or linux 2.6.25).\n");
	goto err;
    }

    /* build feature bitmap */
    for (i = 0; i < sizeof(para_features)/sizeof(para_features[0]); i++) {
	if (kvm_check_extension(xen->kvm, para_features[i].cap))
	    xen->features |= (1 << para_features[i].feature);
    }
    
    /* create vm */
    xen->pg_total &= ~((1<<10)-1);  /* round to 4MB border */
    if (vmcore_init(xen) < 0)
	goto err;
    if (kvm_create_vm(xen->kvm) < 0) {
	logprintf(xen, "kvm_create_vm failed\n");
	goto err;
    }

    if (kvm_set_tss_addr(xen->kvm, xen->pg_total * PAGE_SIZE) < 0) {
	logprintf(xen, "kvm_set_tss_addr failed (%s)\n", strerror(errno));
	goto err;
    }

    if (kvm_register_userspace_phys_mem(xen->kvm, 0, xen->memory,
					xen->pg_total * PAGE_SIZE, 1) < 0) {
	logprintf(xen, "kvm_register_userspace_phys_mem failed\n");
	goto err;
    }

    kvm_create_irqchip(xen->kvm);
    if (!kvm_irqchip_in_kernel(xen->kvm)) {
	logprintf(xen, "kvm_create_irqchip failed\n");
	goto err;
    }

    for (i = 0; i < xen->vcpus; i++) {
	if (kvm_create_vcpu(xen->kvm, i) < 0) {
	    logprintf(xen, "kvm_create_vcpu failed\n");
	    goto err;
	}
	xen->vcpu[i].id = i;
	xen->vcpu[i].vm = xen;
	need_regs(xen->vcpu + i);
	need_sregs(xen->vcpu + i);
    }

    banner_print(xen, "setup");
    sigsetup(xen);
    if (0 != evtchn_init(xen))
	goto err;
    if (0 != xenstore_init_early(xen))
	goto err;
    console_init(xen);

    if (0 != domain_builder(xen)) {
	logprintf(xen, "domain builder failed\n");
	goto err;
    }

    xenstore_init_late(xen);
    setup_cpuid(xen);

    tsc_calibrate(xen, &xen->tsc_shift, &xen->tsc_mul_frac);
    xen->boot = get_systime();

    vmcore_write_headers(xen);
    return 0;

err:
    if (xen->kvm)
	kvm_finalize(xen->kvm);
    if (-1 != xen->corefd)
	unlink(xen->corefile);
    return -1;
}

/* ------------------------------------------------------------------ */

static void *vcpu_thread(void *arg)
{
    struct xenvcpu *vcpu = arg;
    struct xenvm *xen = vcpu->vm;

    sigprocmask(SIG_BLOCK,   &main_signals, NULL);

    pthread_mutex_lock(&xen->biglock);
    d1printf("%s/%d: start\n", __FUNCTION__, vcpu->id);
    vcpu->online = 1;

    setup_regs(vcpu);
    flush_regs(vcpu);
    flush_sregs(vcpu);

    kvm_run(xen->kvm, vcpu->id);

    vcpu->online = 0;
    d1printf("%s/%d: stop\n", __FUNCTION__, vcpu->id);
    pthread_mutex_unlock(&xen->biglock);
    return NULL;
}

int kvmbits_mainloop(struct xenvm *xen)
{
    struct xenvcpu *vcpu = xen->vcpu + 0;
    struct list_head *item;
    struct xenner_callback *cb;
    fd_set rd;
    int rc, max;

    pthread_mutex_lock(&xen->biglock);
    pthread_create(&vcpu->thread, NULL, vcpu_thread, vcpu);

    for (;;) {
	if (termsig)
	    vm_kill(vcpu, "termsig", 0);
	check_events(xen);

	FD_ZERO(&rd);
	max = 0;
	list_for_each(item, &xen->callbacks[XENNER_CB_IOFD]) {
	    cb = list_entry(item, struct xenner_callback, list);
	    FD_SET(cb->handle,&rd);
	    if (max < cb->handle)
		max = cb->handle;
	}

	pthread_mutex_unlock(&xen->biglock);
	rc = select(max+1, &rd, NULL, NULL, NULL);
	pthread_mutex_lock(&xen->biglock);
	
	if (-1 == rc) {
	    if (EINTR == errno) {
		d1printf("%s: select: EINTR (got signal)\n", __FUNCTION__);
		continue;
	    } else {
		d0printf("%s: select: %s\n", __FUNCTION__, strerror(errno));
		vm_kill(vcpu, "select error", 0);
	    }
	}
	list_for_each(item, &xen->callbacks[XENNER_CB_IOFD]) {
	    cb = list_entry(item, struct xenner_callback, list);
	    if (FD_ISSET(cb->handle,&rd))
		cb->callback(xen, cb->handle, cb->cb_data);
	}
    }

    pthread_mutex_unlock(&xen->biglock);
    return 0;
}

int kvmbits_vcpu_up(struct xenvcpu *vcpu)
{
    if (vcpu->thread)
	return -EINVAL;
    pthread_create(&vcpu->thread, NULL, vcpu_thread, vcpu);
    return 0;
}

int kvmbits_vcpu_down(struct xenvcpu *vcpu)
{
    /* FIXME */
    return -ENOSYS;
}
